/*
 * Copyright (C) 2009 Philippe Gerum <rpm@xenomai.org>.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.

 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA.
 *
 *
 * XDDP-based RT/NRT threads communication demo.
 *
 * Real-time Xenomai threads and regular Linux threads may want to
 * exchange data in a way that does not require the former to leave
 * the real-time domain (i.e. secondary mode). Message pipes - as
 * implemented by the RTDM-based XDDP protocol - are provided for this
 * purpose.
 *
 * On the Linux domain side, pseudo-device files named /dev/rtp<minor>
 * give regular POSIX threads access to non real-time communication
 * endpoints, via the standard character-based I/O interface. On the
 * Xenomai domain side, sockets may be bound to XDDP ports, which act
 * as proxies to send and receive data to/from the associated
 * pseudo-device files. Ports and pseudo-device minor numbers are
 * paired, meaning that e.g. port 7 will proxy the traffic for
 * /dev/rtp7. Therefore, port numbers may range from 0 to
 * CONFIG_XENO_OPT_PIPE_NRDEV - 1.
 *
 * All data sent through a bound/connected XDDP socket via sendto(2) or
 * write(2) will be passed to the peer endpoint in the Linux domain,
 * and made available for reading via the standard read(2) system
 * call. Conversely, all data sent using write(2) through the non
 * real-time endpoint will be conveyed to the real-time socket
 * endpoint, and made available to the recvfrom(2) or read(2) system
 * calls.
 *
 * In addition to sending datagrams, real-time threads may stream data
 * in a byte-oriented mode through the proxy as well. This increases
 * the bandwidth and reduces the overhead, when a lot of data has to
 * flow down to the Linux domain, if keeping the message boundaries is
 * not required. The example code below illustrates such use.
 *
 * realtime_thread-------------------------------------->----------+
 *   =>  get socket                                                |
 *   =>  bind socket to port 0                                     v
 *   =>  write scattered traffic to NRT domain via sendto()        |
 *   =>  read traffic from NRT domain via recvfrom()            <--|--+
 *                                                                 |  |
 * regular_thread--------------------------------------------------+  |
 *   =>  open /dev/rtp0                                            |  ^
 *   =>  read traffic from RT domain via read()                    |  |
 *   =>  echo traffic back to RT domain via write()                +--+
 */
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <signal.h>
#include <string.h>
#include <malloc.h>
#include <pthread.h>
#include <fcntl.h>
#include <errno.h>
#include <execinfo.h>
#include <iostream>
// #define EIGEN_DONT_VECTORIZE
#include <eigen3/Eigen/Dense>


pthread_t rt, nrt;

Eigen::MatrixXd A, B, H;
int nrows = 20;

#define XDDP_PORT 0 /* [0..CONFIG-XENO_OPT_PIPE_NRDEV - 1] */

static void warn_upon_switch(int sig)
{
    
    void *bt[32];
    int nentries;
    
    nentries = backtrace(bt, sizeof(bt)/sizeof(bt[0]));
    backtrace_symbols_fd(bt, nentries, fileno(stdout));

    
}

void test_mul(int nrows_, int ncols_)
{
    
    A.setRandom(nrows_, ncols_);
    B.setRandom(ncols_, nrows_);
    H.setZero(nrows_, nrows_);
    
    H = A * B;
    
}

static void fail(const char *reason)
{
    perror(reason);
    exit(EXIT_FAILURE);
}

// static void *realtime_thread(void *arg)
// {
// 
// //     pthread_setmode_np(0, PTHREAD_WARNSW, NULL);
//     
//     struct timespec ts;
// 
//     for (int ncols = 1; ncols < 200; ncols++) {
// //         test_mul(nrows, ncols);
// //         printf("RT: Nrows %d Ncols %d SUCCESS!\n", nrows, ncols);
//         
//         ts.tv_sec = 0;
//         ts.tv_nsec = 10000000; /* 10 ms */
//         clock_nanosleep(CLOCK_REALTIME, 0, &ts, NULL);
//     }
// 
//     return NULL;
// }

static void *regular_thread(void *arg)
{

    for (int ncols = 1; ncols < 200; ncols++) {
        test_mul(nrows, ncols);
        printf("Nrows %d Ncols %d SUCCESS!\n", nrows, ncols);
        usleep(10000);
    }

    return NULL;
}

int main(int argc, char **argv)
{
    
    for (int ncols = 1; ncols < 200; ncols++) {
        test_mul(nrows, ncols);
        printf("Nrows %d Ncols %d SUCCESS!\n", nrows, ncols);
        usleep(10000);
    }
    
    printf("Main mul test SUCCEDED!\n");
    sleep(1);
    
    
    struct sched_param rtparam = { .sched_priority = 42 };
    pthread_attr_t rtattr, regattr;
    sigset_t set;
    int sig;
    
#ifdef __COBALT__
    signal(SIGDEBUG, warn_upon_switch);
#endif

    sigemptyset(&set);
    sigaddset(&set, SIGINT);
    sigaddset(&set, SIGTERM);
    sigaddset(&set, SIGHUP);
    pthread_sigmask(SIG_BLOCK, &set, NULL);

//     pthread_attr_init(&rtattr);
//     pthread_attr_setdetachstate(&rtattr, PTHREAD_CREATE_JOINABLE);
//     pthread_attr_setinheritsched(&rtattr, PTHREAD_EXPLICIT_SCHED);
//     
// #ifdef __COBALT__
//     pthread_attr_setschedpolicy(&rtattr, SCHED_FIFO);
// #else
//     pthread_attr_setschedpolicy(&rtattr, SCHED_OTHER);
// #endif
//     pthread_attr_setschedparam(&rtattr, &rtparam);
// 
//     errno = pthread_create(&rt, &rtattr, &realtime_thread, NULL);
//     if (errno)
//         fail("pthread_create");

    pthread_attr_init(&regattr);
    pthread_attr_setdetachstate(&regattr, PTHREAD_CREATE_JOINABLE);
    pthread_attr_setinheritsched(&regattr, PTHREAD_EXPLICIT_SCHED);
    pthread_attr_setschedpolicy(&regattr, SCHED_OTHER);

    errno = pthread_create(&nrt, &regattr, &regular_thread, NULL);
    if (errno)
        fail("pthread_create");

    sigwait(&set, &sig);
//     pthread_cancel(rt);
    pthread_cancel(nrt);
//     pthread_join(rt, NULL);
    pthread_join(nrt, NULL);

    return 0;
}